"""
    compare
    ~~~~~~~

    Compare objects for django-reversion-compare

    :copyleft: 2012-2016 by the django-reversion-compare team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import logging

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.utils.encoding import force_text
from django.utils.translation import ugettext as _

from reversion import is_registered
from reversion.models import Version
from reversion.revisions import _get_options

logger = logging.getLogger(__name__)


class FieldVersionDoesNotExist:
    """
    Sentinel object to handle missing fields
    """

    def __str__(self):
        return force_text(_("Field didn't exist!"))


DOES_NOT_EXIST = FieldVersionDoesNotExist()


class CompareObject:
    def __init__(self, field, field_name, obj, version_record, follow):
        self.field = field
        self.field_name = field_name
        self.obj = obj
        self.version_record = version_record  # instance of reversion.models.Version()
        self.follow = follow
        # try and get a value, if none punt
        self.compare_foreign_objects_as_id = getattr(settings, "REVERSION_COMPARE_FOREIGN_OBJECTS_AS_ID", False)
        # ignore not registered models
        self.ignore_not_registered = getattr(settings, "REVERSION_COMPARE_IGNORE_NOT_REGISTERED", False)
        if self.compare_foreign_objects_as_id:
            self.value = version_record.field_dict.get(getattr(field, "attname", field_name), DOES_NOT_EXIST)
        else:
            self.value = version_record.field_dict.get(field_name, DOES_NOT_EXIST)

    def _obj_repr(self, obj):
        # FIXME: How to create a better representation of the current value?
        try:
            return force_text(obj)
        except Exception:
            return repr(obj)

    def _choices_repr(self, obj):
        return force_text(dict(self.field.flatchoices).get(obj, obj), strings_only=True)

    def _to_string_ManyToManyField(self):
        return ", ".join(self._obj_repr(item) for item in self.get_many_to_many())

    def _to_string_ForeignKey(self):
        return self._obj_repr(self.get_related())

    def to_string(self):
        internal_type = self.field.get_internal_type()
        func_name = f"_to_string_{internal_type}"
        if hasattr(self, func_name):
            func = getattr(self, func_name)
            return func()

        if self.field.choices:
            return self._choices_repr(self.value)

        if isinstance(self.value, str):
            return self.value
        else:
            return self._obj_repr(self.value)

    def __cmp__(self, other):
        raise NotImplementedError

    def __eq__(self, other):
        if hasattr(self.field, "get_internal_type"):
            assert self.field.get_internal_type() != "ManyToManyField"

        if self.value != other.value:
            return False

        # see - https://hynek.me/articles/hasattr/
        if not self.compare_foreign_objects_as_id:
            internal_type = getattr(self.field, "get_internal_type", None)
            if internal_type is None or internal_type() == "ForeignKey":  # FIXME!
                if self.version_record.field_dict != other.version_record.field_dict:
                    return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def get_object_version(self):
        if hasattr(self.version_record, "_object_version"):
            return getattr(self.version_record, "_object_version")
        else:
            return getattr(self.version_record, "object_version")

    def get_related(self):
        if getattr(self.field, "related_model", None):
            obj = self.get_object_version().object
            try:
                return getattr(obj, self.field.name, None)
            except ObjectDoesNotExist:
                return None

    def get_reverse_foreign_key(self):
        obj = self.get_object_version().object
        if self.field.related_name and hasattr(obj, self.field.related_name):
            if isinstance(self.field, models.fields.related.OneToOneRel):
                try:
                    ids = {force_text(getattr(obj, force_text(self.field.related_name)).pk)}
                except ObjectDoesNotExist:
                    ids = set()
            else:
                # If there is a _ptr this is a multi-inheritance table and inherits from a non-abstract class
                ids = {force_text(v.pk) for v in getattr(obj, force_text(self.field.related_name)).all()}
                if not ids and any([f.name.endswith("_ptr") for f in obj._meta.get_fields()]):
                    # If there is a _ptr this is a multi-inheritance table and inherits from a non-abstract class
                    # lets try and get the parent items associated entries for this field
                    others = self.version_record.revision.version_set.filter(
                        object_id=self.version_record.object_id
                    ).all()
                    for p in others:
                        if hasattr(p, "_object_version"):
                            p_obj = getattr(p, "_object_version").object
                        else:
                            p_obj = getattr(p, "object_version").object
                        if not isinstance(p_obj, type(obj)) and hasattr(p_obj, force_text(self.field.related_name)):
                            ids = {force_text(v.pk) for v in getattr(p_obj, force_text(self.field.related_name)).all()}
        else:
            return {}, {}, []  # TODO: refactor that

        # Get the related model of the current field:
        related_model = self.field.field.model
        return self.get_many_to_something(ids, related_model, is_reverse=True)

    def get_many_to_many(self):
        """
        returns a queryset with all many2many objects
        """
        if self.field.get_internal_type() != "ManyToManyField":  # FIXME!
            return {}, {}, []  # TODO: refactor that
        elif self.value is DOES_NOT_EXIST:
            return {}, {}, []  # TODO: refactor that

        try:
            ids = frozenset(map(force_text, self.value))
        except TypeError:
            # catch errors e.g. produced by taggit's TaggableManager
            logger.exception("Can't collect m2m ids")
            return {}, {}, []  # TODO: refactor that

        # Get the related model of the current field:
        return self.get_many_to_something(ids, self.field.related_model)

    def get_many_to_something(self, target_ids, related_model, is_reverse=False):
        # get instance of reversion.models.Revision():
        # A group of related object versions.
        old_revision = self.version_record.revision

        # Get a queryset with all related objects.
        versions = {
            ver.object_id: ver
            for ver in old_revision.version_set.filter(
                content_type=ContentType.objects.get_for_model(related_model), object_id__in=target_ids
            ).all()
        }

        missing_objects_dict = {}
        deleted = []

        if not self.follow:
            # This models was not registered with follow relations
            # Try to fill missing related objects
            potentially_missing_ids = target_ids.difference(frozenset(versions))
            # logger.debug(
            #     self.field_name,
            #     f"target: {target_ids} - actual: {versions} - missing: {potentially_missing_ids}"
            # )
            if potentially_missing_ids:
                missing_objects_dict = {
                    force_text(rel.pk): rel
                    for rel in related_model.objects.filter(pk__in=potentially_missing_ids).iterator()
                    if is_registered(rel.__class__) or not self.ignore_not_registered
                }

        if is_reverse:
            missing_objects_dict = {
                ver.object_id: ver
                for o in missing_objects_dict.values()
                for ver in Version.objects.get_for_object(o)
                if ver.revision.date_created < old_revision.date_created
            }

            if is_registered(related_model) or not self.ignore_not_registered:
                # shift query to database
                deleted = list(Version.objects.filter(revision=old_revision).get_deleted(related_model))
            else:
                deleted = []

        return versions, missing_objects_dict, deleted

    def get_debug(self):  # pragma: no cover
        if not settings.DEBUG:
            return

        result = [
            f"field..............: {self.field!r}",
            f"field_name.........: {self.field_name!r}",
            "field internal type: %r" % self.field.get_internal_type(),
            "field_dict.........: %s" % repr(self.version_record.field_dict),
            f"obj................: {self.obj!r} (pk: {self.obj.pk}, id: {id(self.obj)})",
            "version............: %r (pk: %s, id: %s)"
            % (self.version_record, self.version_record.pk, id(self.version_record)),
            f"value..............: {self.value!r}",
            "to string..........: %s" % repr(self.to_string()),
            "related............: %s" % repr(self.get_related()),
        ]
        m2m_versions, missing_objects, missing_ids, deleted = self.get_many_to_many()
        if m2m_versions or missing_objects or missing_ids:
            result.append(
                "many-to-many.......: %s" % ", ".join(f"{item} ({item.type})" for item in m2m_versions)
            )

            if missing_objects:
                result.append("missing m2m objects: %s" % repr(missing_objects))
            else:
                result.append("missing m2m objects: (has no)")

            if missing_ids:
                result.append("missing m2m IDs....: %s" % repr(missing_ids))
            else:
                result.append("missing m2m IDs....: (has no)")
        else:
            result.append("many-to-many.......: (has no)")

        return result

    def debug(self):  # pragma: no cover
        if not settings.DEBUG:
            return
        for item in self.get_debug():
            logger.debug(item)


class CompareObjects:
    def __init__(self, field, field_name, obj, version1, version2, is_reversed):
        self.field = field
        self.field_name = field_name
        self.obj = obj

        # is a related field (ForeignKey, ManyToManyField etc.)
        self.is_related = getattr(self.field, "related_model", None) is not None
        self.is_reversed = is_reversed
        if not self.is_related:
            self.follow = None
        elif self.field_name in _get_options(self.obj.__class__).follow:
            self.follow = True
        else:
            self.follow = False

        self.compare_obj1 = CompareObject(field, field_name, obj, version1, self.follow)
        self.compare_obj2 = CompareObject(field, field_name, obj, version2, self.follow)

        self.value1 = self.compare_obj1.value
        self.value2 = self.compare_obj2.value

        self.M2O_CHANGE_INFO = None
        self.M2M_CHANGE_INFO = None

    def changed(self):
        """ return True if at least one field has changed values. """

        info = None
        if hasattr(self.field, "get_internal_type") and self.field.get_internal_type() == "ManyToManyField":
            info = self.get_m2m_change_info()
        elif self.is_reversed:
            info = self.get_m2o_change_info()
        if info:
            keys = (
                "changed_items",
                "removed_items",
                "added_items",
                "removed_missing_objects",
                "added_missing_objects",
                "deleted_items",
            )
            for key in keys:
                if info[key]:
                    return True
            return False

        return self.compare_obj1 != self.compare_obj2

    def to_string(self):
        return self.compare_obj1.to_string(), self.compare_obj2.to_string()

    def get_related(self):
        return self.compare_obj1.get_related(), self.compare_obj2.get_related()

    def get_many_to_many(self):
        return self.compare_obj1.get_many_to_many(), self.compare_obj2.get_many_to_many()

    def get_reverse_foreign_key(self):
        return self.compare_obj1.get_reverse_foreign_key(), self.compare_obj2.get_reverse_foreign_key()

    def get_m2o_change_info(self):
        if self.M2O_CHANGE_INFO is not None:
            return self.M2O_CHANGE_INFO

        m2o_data1, m2o_data2 = self.get_reverse_foreign_key()

        self.M2O_CHANGE_INFO = self.get_m2s_change_info(m2o_data1, m2o_data2)
        return self.M2O_CHANGE_INFO

    def get_m2m_change_info(self):
        if self.M2M_CHANGE_INFO is not None:
            return self.M2M_CHANGE_INFO

        m2m_data1, m2m_data2 = self.get_many_to_many()

        self.M2M_CHANGE_INFO = self.get_m2s_change_info(m2m_data1, m2m_data2)
        return self.M2M_CHANGE_INFO

    # Abstract Many-to-Something (either -many or -one) as both
    # many2many and many2one relationships looks the same from the referred object.
    def get_m2s_change_info(self, obj1_data, obj2_data):

        result_dict1, missing_objects_dict1, deleted1 = obj1_data
        result_dict2, missing_objects_dict2, deleted2 = obj2_data

        # Create same_items, removed_items, added_items with related m2m items
        changed_items = []
        removed_items = []
        added_items = []
        same_items = []

        same_missing_objects_dict = {k: v for k, v in missing_objects_dict1.items() if k in missing_objects_dict2}
        removed_missing_objects_dict = {
            k: v for k, v in missing_objects_dict1.items() if k not in missing_objects_dict2
        }
        added_missing_objects_dict = {k: v for k, v in missing_objects_dict2.items() if k not in missing_objects_dict1}

        # logger.debug("same_missing_objects: %s", same_missing_objects_dict)
        # logger.debug("removed_missing_objects: %s", removed_missing_objects_dict)
        # logger.debug("added_missing_objects: %s", added_missing_objects_dict)

        for primary_key in set().union(result_dict1.keys(), result_dict2.keys()):
            if primary_key in result_dict1:
                version1 = result_dict1[primary_key]
            else:
                version1 = None

            if primary_key in result_dict2:
                version2 = result_dict2[primary_key]
            else:
                version2 = None

            # logger.debug("%r - %r - %r", primary_key, version1, version2)

            if version1 is not None and version2 is not None:
                # In both -> version changed or the same
                if version1.serialized_data == version2.serialized_data:
                    # logger.debug("same item: %s", version1)
                    same_items.append(version1)
                else:
                    changed_items.append((version1, version2))
            elif version1 is not None and version2 is None:
                # In 1 but not in 2 -> removed
                # logger.debug("%s %s", primary_key, missing_objects_dict2)
                # logger.debug("%s %s", repr(primary_key), repr(missing_objects_dict2))
                if primary_key in added_missing_objects_dict:
                    added_missing_objects_dict.pop(primary_key)
                    same_missing_objects_dict[primary_key] = missing_objects_dict2[primary_key]
                    continue
                removed_items.append(version1)
            elif version1 is None and version2 is not None:
                # In 2 but not in 1 -> added
                # logger.debug("added: %s", version2)
                added_items.append(version2)
            else:
                raise RuntimeError()

        # In Place Sorting of Lists (exclude changed since its a tuple)
        removed_items.sort(key=lambda item: force_text(item))
        added_items.sort(key=lambda item: force_text(item))
        same_items.sort(key=lambda item: force_text(item))
        deleted1.sort(key=lambda item: force_text(item))
        same_missing_objects = sorted(same_missing_objects_dict.values(), key=lambda item: force_text(item))
        removed_missing_objects = sorted(removed_missing_objects_dict.values(), key=lambda item: force_text(item))
        added_missing_objects = sorted(added_missing_objects_dict.values(), key=lambda item: force_text(item))

        return {
            "changed_items": changed_items,
            "removed_items": removed_items,
            "added_items": added_items,
            "same_items": same_items,
            "same_missing_objects": same_missing_objects,
            "removed_missing_objects": removed_missing_objects,
            "added_missing_objects": added_missing_objects,
            "deleted_items": deleted1,
        }

    def debug(self):  # pragma: no cover
        if not settings.DEBUG:
            return
        logger.debug("_______________________________")
        logger.debug(" *** CompareObjects debug: ***")
        logger.debug("changed: %s", self.changed())
        logger.debug("follow: %s", self.follow)

        debug1 = self.compare_obj1.get_debug()
        debug2 = self.compare_obj2.get_debug()
        debug_set1 = set(debug1)
        debug_set2 = set(debug2)

        logger.debug(" *** same attributes/values in obj1 and obj2: ***")
        intersection = debug_set1.intersection(debug_set2)
        for item in debug1:
            if item in intersection:
                logger.debug(item)

        logger.debug(" -" * 40)
        logger.debug(" *** unique attributes/values from obj1: ***")
        difference = debug_set1.difference(debug_set2)
        for item in debug1:
            if item in difference:
                logger.debug(item)

        logger.debug(" -" * 40)
        logger.debug(" *** unique attributes/values from obj2: ***")
        difference = debug_set2.difference(debug_set1)
        for item in debug2:
            if item in difference:
                logger.debug(item)

        logger.debug("-" * 79)



from pathlib import Path

import reversion_compare
from creole.setup_utils import assert_rst_readme, update_rst_readme
from poetry_publish.publish import poetry_publish
from poetry_publish.utils.subprocess_utils import verbose_check_call

PACKAGE_ROOT = Path(reversion_compare.__file__).parent.parent


def update_readme():
    return update_rst_readme(
        package_root=PACKAGE_ROOT,
        filename='README.creole'
    )


def publish():
    """
        Publish python-creole to PyPi
        Call this via:
            $ poetry run publish
    """
    # don't publish if README is not up-to-date:
    assert_rst_readme(package_root=PACKAGE_ROOT, filename='README.creole')

    # don't publish if code style wrong:
    verbose_check_call('make', 'fix-code-style')

    poetry_publish(
        package_root=PACKAGE_ROOT,
        version=reversion_compare.__version__,
        creole_readme=True  # don't publish if README.rst is not up-to-date
    )



"""
    mixin
    ~~~~~

    Mixins for views (admin and cbv) for django-reversion-compare

    :copyleft: 2012-2015 by the django-reversion-compare team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""
import django
from django.db import models
from django.template.loader import render_to_string
from django.utils.encoding import force_text

from reversion_compare.compare import CompareObjects
from reversion_compare.helpers import html_diff


class CompareMixin:
    """A mixin to add comparison capabilities to your views"""

    # list/tuple of field names for the compare view. Set to None for all existing fields
    compare_fields = None

    # list/tuple of field names to exclude from compare view.
    compare_exclude = None

    # sort from new to old as default, see: https://github.com/etianen/django-reversion/issues/77
    history_latest_first = True

    def _order_version_queryset(self, queryset):
        """Applies the correct ordering to the given version queryset."""
        if self.history_latest_first:
            return queryset.order_by("-pk")
        return queryset.order_by("pk")

    def _get_compare(self, obj_compare):
        """
        Call the methods to create the compare html part.
        Try:
            1. name scheme: "compare_%s" % field_name
            2. name scheme: "compare_%s" % field.get_internal_type()
            3. Fallback to: self.fallback_compare()
        """

        def _get_compare_func(suffix):
            # logger.debug("func_name: %s", func_name)
            func_name = f"compare_{suffix}"
            if hasattr(self, func_name):
                func = getattr(self, func_name)
                if callable(func):
                    return func

        # Try method in the name scheme: "compare_%s" % field_name
        func = _get_compare_func(obj_compare.field_name)
        if func is not None:
            html = func(obj_compare)
            return html

        # Determine if its a reverse field
        if obj_compare.field in self.reverse_fields:
            func = _get_compare_func("ManyToOneRel")
            if func is not None:
                html = func(obj_compare)
                return html

        # Try method in the name scheme: "compare_%s" % field.get_internal_type()
        internal_type = obj_compare.field.get_internal_type()
        func = _get_compare_func(internal_type)
        if func is not None:
            html = func(obj_compare)
            return html

        # Fallback to self.fallback_compare()
        html = self.fallback_compare(obj_compare)
        return html

    def compare(self, obj, version1, version2):
        """
        Create a generic html diff from the obj between version1 and version2:

            A diff of every changes field values.

        This method should be overwritten, to create a nice diff view
        coordinated with the model.
        """
        diff = []

        # Create a list of all normal fields and append many-to-many fields
        fields = [field for field in obj._meta.fields]
        concrete_model = obj._meta.concrete_model
        fields += concrete_model._meta.many_to_many

        # This gathers the related reverse ForeignKey fields, so we can do ManyToOne compares
        if django.VERSION < (1, 10):
            # From: http://stackoverflow.com/questions/19512187/django-list-all-reverse-relations-of-a-model
            self.reverse_fields = []
            for field_name in obj._meta.get_all_field_names():
                f = getattr(obj._meta.get_field_by_name(field_name)[0], "field", None)
                if isinstance(f, models.ForeignKey) and f not in fields:
                    self.reverse_fields.append(f.rel)
        else:
            # django >= v1.10
            self.reverse_fields = []
            for field in obj._meta.get_fields(include_hidden=True):
                f = getattr(field, "field", None)
                if isinstance(f, models.ForeignKey) and f not in fields:
                    self.reverse_fields.append(f.remote_field)

        fields += self.reverse_fields

        has_unfollowed_fields = False

        for field in fields:
            # logger.debug("%s %s %s", field, field.db_type, field.get_internal_type())
            try:
                field_name = field.name
            except BaseException:
                # is a reverse FK field
                field_name = field.field_name

            if self.compare_fields and field_name not in self.compare_fields:
                continue
            if self.compare_exclude and field_name in self.compare_exclude:
                continue

            is_reversed = field in self.reverse_fields
            obj_compare = CompareObjects(field, field_name, obj, version1, version2, is_reversed)
            # obj_compare.debug()

            is_related = obj_compare.is_related
            follow = obj_compare.follow
            if is_related and not follow:
                has_unfollowed_fields = True

            if not obj_compare.changed():
                # Skip all fields that aren't changed
                continue

            html = self._get_compare(obj_compare)
            diff.append({"field": field, "is_related": is_related, "follow": follow, "diff": html})

        return diff, has_unfollowed_fields

    def fallback_compare(self, obj_compare):
        """
        Simply create a html diff from the repr() result.
        Used for every field which has no own compare method.
        """
        value1, value2 = obj_compare.to_string()
        html = html_diff(value1, value2)
        return html


class CompareMethodsMixin:
    """A mixin to add prepared compare methods."""

    def generic_add_remove(self, raw_value1, raw_value2, value1, value2):
        if raw_value1 is None:
            # a new values was added:
            context = {"value": value2}
            return render_to_string("reversion-compare/compare_generic_add.html", context)
        elif raw_value2 is None:
            # the existing value was removed:
            context = {"value": value1}
            return render_to_string("reversion-compare/compare_generic_remove.html", context)
        else:
            html = html_diff(value1, value2)
            return html

    def compare_ForeignKey(self, obj_compare):
        related1, related2 = obj_compare.get_related()
        # obj_compare.debug()
        value1, value2 = force_text(related1), force_text(related2)
        return self.generic_add_remove(related1, related2, value1, value2)

    def simple_compare_ManyToManyField(self, obj_compare):
        """ comma separated list of all m2m objects """
        m2m1, m2m2 = obj_compare.get_many_to_many()
        old = ", ".join(force_text(item) for item in m2m1)
        new = ", ".join(force_text(item) for item in m2m2)
        html = html_diff(old, new)
        return html

    def compare_ManyToOneRel(self, obj_compare):
        change_info = obj_compare.get_m2o_change_info()
        context = {"change_info": change_info}
        return render_to_string("reversion-compare/compare_generic_many_to_many.html", context)

    def compare_ManyToManyField(self, obj_compare):
        """ create a table for m2m compare """
        change_info = obj_compare.get_m2m_change_info()
        context = {"change_info": change_info}
        return render_to_string("reversion-compare/compare_generic_many_to_many.html", context)

    # compare_ManyToManyField = simple_compare_ManyToManyField

    def compare_FileField(self, obj_compare):
        value1 = obj_compare.value1
        value2 = obj_compare.value2

        # FIXME: Needed to not get 'The 'file' attribute has no file associated with it.'
        if value1:
            value1 = value1.url
        else:
            value1 = None

        if value2:
            value2 = value2.url
        else:
            value2 = None

        return self.generic_add_remove(value1, value2, value1, value2)

    def compare_DateTimeField(self, obj_compare):
        """ compare all model datetime field in ISO format """
        context = {"date1": obj_compare.value1, "date2": obj_compare.value2}
        return render_to_string("reversion-compare/compare_DateTimeField.html", context)

    def compare_BooleanField(self, obj_compare):
        """ compare booleans as a complete field, rather than as a string """
        context = {"bool1": obj_compare.value1, "bool2": obj_compare.value2}
        return render_to_string("reversion-compare/compare_BooleanField.html", context)

    compare_NullBooleanField = compare_BooleanField



from django.http import Http404
from django.shortcuts import get_object_or_404
from django.views.generic.detail import DetailView

from reversion.models import Version
from reversion_compare.forms import SelectDiffForm
from reversion_compare.mixins import CompareMethodsMixin, CompareMixin


class HistoryCompareDetailView(CompareMixin, CompareMethodsMixin, DetailView):
    """This class can be used to add a non-admin view for comparing your object's versions.

    You can use it just like a normal DetailView:

    Inherit from it in your class and add a model (or queryset), for example:

    class SimpleModelHistoryCompareView(HistoryCompareDetailView):
        model = SimpleModel

    and assign that CBV to a url:

    url(r'^test_view/(?P<pk>\d+)$', views.SimpleModelHistoryCompareView.as_view() ), # noqa flake8

    Last step, you need to create a template to display both the version select form and
    the changes part (if the form is submitted). An example template is the following:

    <style type="text/css">
    /* minimal style for the diffs */
    del, ins {
        color: #000;
        text-decoration: none;
    }
    del { background-color: #ffe6e6; }
    ins { background-color: #e6ffe6; }
    sup.follow { color: #5555ff; }
    </style>


    {% include "reversion-compare/action_list_partial.html"  %}
    {% include "reversion-compare/compare_partial.html"  %}
    {% include "reversion-compare/compare_links_partial.html"  %}


    Beyond the styling, you should include
    - reversion-compare/action_list_partial.html partial template to display the version select form
    - reversion-compare/compare_partial.html partial template to display the actual version
    - reversion-compare/compare_links_partial.html to include previous/next comparison links

    If you want more control on the appearence of your templates you can check these partials
    to understand how the availabble context variables are used.
    """

    def _get_action_list(self,):
        action_list = [
            {"version": version, "revision": version.revision}
            for version in self._order_version_queryset(
                Version.objects.get_for_object(self.get_object()).select_related("revision__user")
            )
        ]
        return action_list

    def get_context_data(self, **kwargs):
        context = super().get_context_data()
        action_list = self._get_action_list()

        if len(action_list) < 2:
            # Less than two history items aren't enough to compare ;)
            comparable = False
        else:
            comparable = True
            # for pre selecting the compare radio buttons depend on the ordering:
            if self.history_latest_first:
                action_list[0]["first"] = True
                action_list[1]["second"] = True
            else:
                action_list[-1]["first"] = True
                action_list[-2]["second"] = True

        if self.request.GET:
            form = SelectDiffForm(self.request.GET)
            if not form.is_valid():
                msg = "Wrong version IDs."
                raise Http404(msg)

            version_id1 = form.cleaned_data["version_id1"]
            version_id2 = form.cleaned_data["version_id2"]

            if version_id1 > version_id2:
                # Compare always the newest one (#2) with the older one (#1)
                version_id1, version_id2 = version_id2, version_id1

            obj = self.get_object()
            queryset = Version.objects.get_for_object(obj)
            version1 = get_object_or_404(queryset, pk=version_id1)
            version2 = get_object_or_404(queryset, pk=version_id2)

            next_version = queryset.filter(pk__gt=version_id2).last()
            prev_version = queryset.filter(pk__lt=version_id1).first()

            compare_data, has_unfollowed_fields = self.compare(obj, version1, version2)

            context.update(
                {
                    "compare_data": compare_data,
                    "has_unfollowed_fields": has_unfollowed_fields,
                    "version1": version1,
                    "version2": version2,
                }
            )

            if next_version:
                next_url = f"?version_id1={version2.id:d}&version_id2={next_version.id:d}"
                context.update({"next_url": next_url})
            if prev_version:
                prev_url = f"?version_id1={prev_version.id:d}&version_id2={version1.id:d}"
                context.update({"prev_url": prev_url})

        # Compile the context.
        context.update({"action_list": action_list, "comparable": comparable, "compare_view": True})
        return context



"""
    django-reversion helpers
    ~~~~~~~~~~~~~~~~~~~~~~~~

    A number of useful helper functions to automate common tasks.

    Used google-diff-match-patch [1] if installed, fallback to difflib.
    For installing use e.g. the unofficial package:

        pip install diff-match-patch

    [1] http://code.google.com/p/google-diff-match-patch/

    :copyleft: 2012-2015 by the django-reversion-compare team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""

import difflib
import logging

from django.contrib import admin
from django.contrib.admin.sites import NotRegistered
from django.utils.encoding import force_text
from django.utils.html import escape
from django.utils.safestring import mark_safe

logger = logging.getLogger(__name__)


try:
    # http://code.google.com/p/google-diff-match-patch/
    from diff_match_patch import diff_match_patch
except ImportError:
    dmp = None
else:
    dmp = diff_match_patch()


def highlight_diff(diff_text):
    """
    Simple highlight a diff text in the way pygments do it ;)
    """
    html = ['<pre class="highlight">']
    for line in diff_text.splitlines():
        line = escape(line)
        if line.startswith("+"):
            line = f"<ins>{line}</ins>"
        elif line.startswith("-"):
            line = f"<del>{line}</del>"

        html.append(line)
    html.append("</pre>")
    html = "\n".join(html)

    return html


SEMANTIC = 1
EFFICIENCY = 2

# Change from ndiff to unified_diff if old/new values are more than X lines:
LINE_COUNT_4_UNIFIED_DIFF = 4


def unified_diff(a, b, n=3, lineterm="\n"):
    r"""
    simmilar to the original difflib.unified_diff except:
        - no fromfile/tofile and no fromfiledate/tofiledate info lines
        - newline before diff control lines and not after

    Example:

    >>> for line in unified_diff('one two three four'.split(),
    ...             'zero one tree four'.split(), lineterm=''):
    ...     print(line)
    @@ -1,4 +1,4 @@
    +zero
     one
    -two
    -three
    +tree
     four
    """
    started = False
    for group in difflib.SequenceMatcher(None, a, b).get_grouped_opcodes(n):
        first, last = group[0], group[-1]
        file1_range = difflib._format_range_unified(first[1], last[2])
        file2_range = difflib._format_range_unified(first[3], last[4])

        if not started:
            started = True
            yield f"@@ -{file1_range} +{file2_range} @@"
        else:
            yield f"{lineterm}@@ -{file1_range} +{file2_range} @@"

        for tag, i1, i2, j1, j2 in group:
            if tag == "equal":
                for line in a[i1:i2]:
                    yield " " + line
                continue
            if tag in ("replace", "delete"):
                for line in a[i1:i2]:
                    yield "-" + line
            if tag in ("replace", "insert"):
                for line in b[j1:j2]:
                    yield "+" + line


def html_diff(value1, value2, cleanup=SEMANTIC):
    """
    Generates a diff used google-diff-match-patch is exist or ndiff as fallback

    The cleanup parameter can be SEMANTIC, EFFICIENCY or None to clean up the diff
    for greater human readibility.
    """
    value1 = force_text(value1)
    value2 = force_text(value2)
    if dmp is not None:
        # Generate the diff with google-diff-match-patch
        diff = dmp.diff_main(value1, value2)
        if cleanup == SEMANTIC:
            dmp.diff_cleanupSemantic(diff)
        elif cleanup == EFFICIENCY:
            dmp.diff_cleanupEfficiency(diff)
        elif cleanup is not None:
            raise ValueError("cleanup parameter should be one of SEMANTIC, EFFICIENCY or None.")
        html = dmp.diff_prettyHtml(diff)
        html = html.replace("&para;<br>", "</br>")  # IMHO mark paragraphs are needlessly
    else:
        # fallback: use built-in difflib
        value1 = value1.splitlines()
        value2 = value2.splitlines()

        if len(value1) > LINE_COUNT_4_UNIFIED_DIFF or len(value2) > LINE_COUNT_4_UNIFIED_DIFF:
            diff = unified_diff(value1, value2, n=2)
        else:
            diff = difflib.ndiff(value1, value2)

        diff_text = "\n".join(diff)
        html = highlight_diff(diff_text)

    html = mark_safe(html)

    return html


def compare_queryset(first, second):
    """
    Simple compare two querysets (used for many-to-many field compare)
    XXX: resort results?
    """
    result = []
    for item in set(first).union(set(second)):
        if item not in first:  # item was inserted
            item.insert = True
        elif item not in second:  # item was deleted
            item.delete = True
        result.append(item)
    return result


def patch_admin(model, admin_site=None, AdminClass=None, skip_non_revision=False):
    """
    Enables version control with full admin integration for a model that has
    already been registered with the django admin site.

    This is excellent for adding version control to existing Django contrib
    applications.

    :param skip_non_revision: If ==True: Skip models that are not register with ModelAdmin
    """
    admin_site = admin_site or admin.site
    try:
        ModelAdmin = admin_site._registry[model].__class__
    except KeyError:
        raise NotRegistered(f"The model {model} has not been registered with the admin site.")

    if skip_non_revision:
        if not hasattr(ModelAdmin, "object_history_template"):
            logger.info(
                "Skip activate compare admin, because model %r is not registered with revision manager."
                % model._meta.object_name
            )
        return

    # Unregister existing admin class.
    admin_site.unregister(model)

    # Register patched admin class.
    if not AdminClass:
        from reversion_compare.admin import CompareVersionAdmin

        class PatchedModelAdmin(CompareVersionAdmin, ModelAdmin):
            pass

    else:

        class PatchedModelAdmin(AdminClass, ModelAdmin):
            pass

    admin_site.register(model, PatchedModelAdmin)


if __name__ == "__main__":
    import doctest

    print(
        doctest.testmod(
            verbose=False
            # verbose=True
        )
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['reversion_compare']

package_data = \
{'': ['*'],
 'reversion_compare': ['locale/de/LC_MESSAGES/*',
                       'locale/el/LC_MESSAGES/*',
                       'locale/fi/LC_MESSAGES/*',
                       'locale/fr/LC_MESSAGES/*',
                       'locale/nl/LC_MESSAGES/*',
                       'locale/pl/LC_MESSAGES/*',
                       'templates/reversion-compare/*']}

install_requires = \
['django-reversion', 'django>=2.2']

entry_points = \
{'console_scripts': ['publish = reversion_compare.publish:publish',
                     'run_testserver = '
                     'reversion_compare_tests.manage:start_test_server',
                     'update_rst_readme = '
                     'reversion_compare.publish:update_readme']}

setup_kwargs = {
    'name': 'django-reversion-compare',
    'version': '0.10.0',
    'description': 'history compare for django-reversion',
    'long_description': '========================\ndjango-reversion-compare\n========================\n\n**django-reversion-compare** is an extension to `django-reversion <https://github.com/etianen/django-reversion/>`_ that provides a history compare view to compare two versions of a model which is under reversion.\n\nComparing model versions is not a easy task. Maybe there are different view how this should looks like.\nThis project will gives you a generic way to see whats has been changed.\n\nMany parts are customizable by overwrite methods or subclassing, see above.\n\n+--------------------------------------+--------------------------------------------------------------------+\n| |Build Status on github|             | `github.com/jedie/django-reversion-compare/actions`_               |\n+--------------------------------------+--------------------------------------------------------------------+\n| |Build Status on travis-ci.org|      | `travis-ci.org/jedie/django-reversion-compare`_                    |\n+--------------------------------------+--------------------------------------------------------------------+\n| |Coverage Status on coveralls.io|    | `coveralls.io/r/jedie/django-reversion-compare`_                   |\n+--------------------------------------+--------------------------------------------------------------------+\n| |Coverage Status on codecov.io|      | `codecov.io/gh/jedie/django-reversion-compare`_                    |\n+--------------------------------------+--------------------------------------------------------------------+\n| |Requirements Status on requires.io| | `requires.io/github/jedie/django-reversion-compare/requirements/`_ |\n+--------------------------------------+--------------------------------------------------------------------+\n\n.. |Build Status on github| image:: https://github.com/jedie/django-reversion-compare/workflows/test/badge.svg?branch=master\n.. _github.com/jedie/django-reversion-compare/actions: https://github.com/jedie/django-reversion-compare/actions\n.. |Build Status on travis-ci.org| image:: https://travis-ci.org/jedie/django-reversion-compare.svg\n.. _travis-ci.org/jedie/django-reversion-compare: https://travis-ci.org/jedie/django-reversion-compare/\n.. |Coverage Status on coveralls.io| image:: https://coveralls.io/repos/jedie/django-reversion-compare/badge.svg\n.. _coveralls.io/r/jedie/django-reversion-compare: https://coveralls.io/r/jedie/django-reversion-compare\n.. |Coverage Status on codecov.io| image:: https://codecov.io/gh/jedie/django-reversion-compare/branch/master/graph/badge.svg\n.. _codecov.io/gh/jedie/django-reversion-compare: https://codecov.io/gh/jedie/django-reversion-compare\n.. |Requirements Status on requires.io| image:: https://requires.io/github/jedie/django-reversion-compare/requirements.svg\n.. _requires.io/github/jedie/django-reversion-compare/requirements/: https://requires.io/github/jedie/django-reversion-compare/requirements/\n\n------------\nInstallation\n------------\n\nJust use:\n\n::\n\n    pip install django-reversion-compare\n\nOptionally you can install `google-diff-match-patch <https://code.google.com/p/google-diff-match-patch/>`_, otherwise difflib would be used. The easiest way is to use the unofficial package `diff-match-patch <http://pypi.python.org/pypi/diff-match-patch/>`_, e.g.:\n\n::\n\n    pip install diff-match-patch\n\nSetup\n=====\n\nAdd **reversion_compare** to **INSTALLED_APPS** in your settings.py, e.g.:\n\n::\n\n    INSTALLED_APPS = (\n        \'django...\',\n        ...\n        \'reversion\', # https://github.com/etianen/django-reversion\n        \'reversion_compare\', # https://github.com/jedie/django-reversion-compare\n        ...\n    )\n    \n    # Add reversion models to admin interface:\n    ADD_REVERSION_ADMIN=True\n    # optional settings:\n    REVERSION_COMPARE_FOREIGN_OBJECTS_AS_ID=False\n    REVERSION_COMPARE_IGNORE_NOT_REGISTERED=False\n\nUsage\n=====\n\nInherit from **CompareVersionAdmin** instead of **VersionAdmin** to get the comparison feature.\n\nadmin.py e.g.:\n\n::\n\n    from django.contrib import admin\n    from reversion_compare.admin import CompareVersionAdmin\n    \n    from my_app.models import ExampleModel\n    \n    @admin.register(ExampleModel)\n    class ExampleModelAdmin(CompareVersionAdmin):\n        pass\n\nIf you\'re using an existing third party app, then you can add patch django-reversion-compare into\nits admin class by using the **reversion_compare.helpers.patch_admin()** method. For example, to add\nversion control to the built-in User model:\n\n::\n\n    from reversion_compare.helpers import patch_admin\n    \n    patch_admin(User)\n\ne.g.: Add django-cms Page model:\n\n::\n\n    from cms.models.pagemodel import Page\n    from reversion_compare.helpers import patch_admin\n    \n    \n    # Patch django-cms Page Model to add reversion-compare functionality:\n    patch_admin(Page)\n\nCustomize\n=========\n\nIt\'s possible to change the look for every field or for a entire field type.\nYou must only define a methods to your admin class with this name scheme:\n\n*  ``"compare_%s" % field_name`` \n\n*  ``"compare_%s" % field.get_internal_type()`` \n\nIf there is no method with this name scheme, the ``fallback_compare()`` method will be used.\n\nAn example for specifying a compare method for a model field by name:\n\n::\n\n    class YourAdmin(CompareVersionAdmin):\n        def compare_foo_bar(self, obj_compare):\n            """ compare the foo_bar model field """\n            return "%r <-> %r" % (obj_compare.value1, obj_compare.value2)\n\nand example using **patch_admin** with custom version admin class:\n\n::\n\n    patch_admin(User, AdminClass=YourAdmin)\n\n----------------\nClass Based View\n----------------\n\nBeyond the Admin views, you can also create a Class Based View for displaying and comparing version\ndifferences. This is a single class-based-view that either displays the list of versions to select\nfor an object or displays both the versions **and** their differences (if the versions to be compared\nhave been selected). This class can be used just like a normal DetailView:\n\nInherit from it in your class and add a model (or queryset), for example:\n\n::\n\n    from reversion_compare.views import HistoryCompareDetailView\n    \n    class SimpleModelHistoryCompareView(HistoryCompareDetailView):\n        model = SimpleModel\n\nThen, assign that CBV to a url, for example:\n\n::\n\n    url(r\'^test_view/(?P<pk>\\d+)$\', views.SimpleModelHistoryCompareView.as_view() ),\n\nLast step, you need to create a template to display both the version select form and\nthe changes part (if the form is submitted). An example template is the following:\n\n::\n\n    <style type="text/css">\n    /* minimal style for the diffs */\n    del, ins {\n        color: #000;\n        text-decoration: none;\n    }\n    del { background-color: #ffe6e6; }\n    ins { background-color: #e6ffe6; }\n    sup.follow { color: #5555ff; }\n    </style>\n    \n    {% include "reversion-compare/action_list_partial.html"  %}\n    {% if request.GET.version_id1 %}\n        {% include "reversion-compare/compare_partial.html"  %}\n        {% include "reversion-compare/compare_links_partial.html"  %}\n    {% endif %}\n\nBeyond the styling, you should include:\n\n* reversion-compare/action_list_partial.html partial template to display the version select form\n\n* reversion-compare/compare_partial.html partial template to display the actual version\n\n* reversion-compare/compare_links_partial.html to include previous/next comparison links\n\ncompare_partial.html and compare_links_partial.html will show the compare-related information\nso it\'s better to display them only when the select-versions-tocompare-form has been submitted.\nIf you want more control on the appearence of your templates you can check the above partials\nto understand how the availabble context variables are used and override them completely.\n\n-----------\nScreenshots\n-----------\n\nHere some screenshots of django-reversion-compare:\n\n----\n\nHow to select the versions to compare:\n\n|django-reversion-compare_v0_1_0-01.png|\n\n.. |django-reversion-compare_v0_1_0-01.png| image:: https://raw.githubusercontent.com/jedie/jedie.github.io/master/screenshots/django-reversion-compare/20120508_django-reversion-compare_v0_1_0-01.png\n\n----\n\nfrom **v0.1.0**: DateTimeField compare (last update), TextField compare (content) with small changes and a ForeignKey compare (child model instance was added):\n\n|django-reversion-compare_v0_1_0-02.png|\n\n.. |django-reversion-compare_v0_1_0-02.png| image:: https://raw.githubusercontent.com/jedie/jedie.github.io/master/screenshots/django-reversion-compare/20120508_django-reversion-compare_v0_1_0-02.png\n\n----\n\nfrom **v0.1.0**: Same as above, but the are more lines changed in TextField and the ForeignKey relation was removed:\n\n|django-reversion-compare_v0_1_0-03.png|\n\n.. |django-reversion-compare_v0_1_0-03.png| image:: https://raw.githubusercontent.com/jedie/jedie.github.io/master/screenshots/django-reversion-compare/20120508_django-reversion-compare_v0_1_0-03.png\n\n----\n\nExample screenshot from **v0.3.0**: a many-to-many field compare (friends, hobbies):\n\n|django-reversion-compare_v0_3_0-01.png|\n\n.. |django-reversion-compare_v0_3_0-01.png| image:: https://raw.githubusercontent.com/jedie/jedie.github.io/master/screenshots/django-reversion-compare/20120516_django-reversion-compare_v0_3_0-01.png\n\n* In the first line, the m2m object has been changed.\n\n* line 2: A m2m object was deleted\n\n* line 3: A m2m object was removed from this entry (but not deleted)\n\n* line 4: This m2m object has not changed\n\n----------------------------\ncreate developer environment\n----------------------------\n\ne.g.:\n\n::\n\n    # Clone project (Use your fork SSH url!):\n    ~$ git clone https://github.com/jedie/django-reversion-compare.git\n    ~$ cd django-reversion-compare\n    ~/django-reversion-compare$ make install\n    ~/django-reversion-compare$ make\n    help                 List all commands\n    install-poetry       install or update poetry\n    install              install reversion_compare via poetry\n    lint                 Run code formatters and linter\n    fix-code-style       Fix code formatting\n    tox-listenvs         List all tox test environments\n    tox                  Run pytest via tox with all environments\n    tox-py36             Run pytest via tox with *python v3.6*\n    tox-py37             Run pytest via tox with *python v3.7*\n    tox-py38             Run pytest via tox with *python v3.8*\n    pytest               Run pytest\n    update-rst-readme    update README.rst from README.reversion_compare\n    publish              Release new version to PyPi\n    run-test-server      Start Django dev server with the test project\n\nHelpful for writing and debugging unittests is to run a local test server with the same data.\ne.g.:\n\n::\n\n    ~/django-reversion-compare$ make run-test-server\n\n**migration** will be run and a superuser will be created. Username: **test** Password: **12345678**\n\nCall manage commands from test project, e.g.:\n\n::\n\n    ~/django-reversion-compare$ poetry shell\n    django-reversion-compare-foobar-py3.6) ~/django-reversion-compare$ ./reversion_compare_tests/manage.py --help\n    ...\n\n---------------------\nVersion compatibility\n---------------------\n\n+-------------------+------------------+--------------------+------------------------------------------------+\n| Reversion-Compare | django-reversion | Django             | Python                                         |\n+===================+==================+====================+================================================+\n| >=v0.10.0         | v3.0             | v2.2, v3.0         | v3.6, v3.7, v3.8, pypy3                        |\n+-------------------+------------------+--------------------+------------------------------------------------+\n| >=v0.9.0          | v2.0             | v2.2, v3.0         | v3.6, v3.7, v3.8, pypy3                        |\n+-------------------+------------------+--------------------+------------------------------------------------+\n| >=v0.8.6          | v2.0             | v1.11, v2.0        | v3.5, v3.6, v3.7, pypy3                        |\n+-------------------+------------------+--------------------+------------------------------------------------+\n| >=v0.8.4          | v2.0             | v1.8, v1.11, v2.0  | v3.5, v3.6, pypy3                              |\n+-------------------+------------------+--------------------+------------------------------------------------+\n| >=v0.8.3          | v2.0             | v1.8, v1.11        | v3.5, v3.6, pypy3                              |\n+-------------------+------------------+--------------------+------------------------------------------------+\n| v0.8.x            | v2.0             | v1.8, v1.10, v1.11 | v2.7, v3.4, v3.5, v3.6 (only with Django 1.11) |\n+-------------------+------------------+--------------------+------------------------------------------------+\n| >=v0.7.2          | v2.0             | v1.8, v1.9, v1.10  | v2.7, v3.4, v3.5                               |\n+-------------------+------------------+--------------------+------------------------------------------------+\n| v0.7.x            | v2.0             | v1.8, v1.9         | v2.7, v3.4, v3.5                               |\n+-------------------+------------------+--------------------+------------------------------------------------+\n| v0.6.x            | v1.9, v1.10      | v1.8, v1.9         | v2.7, v3.4, v3.5                               |\n+-------------------+------------------+--------------------+------------------------------------------------+\n| >=v0.5.2          | v1.9             | v1.7, v1.8         | v2.7, v3.4                                     |\n+-------------------+------------------+--------------------+------------------------------------------------+\n| >=v0.4            | v1.8             | v1.7               | v2.7, v3.4                                     |\n+-------------------+------------------+--------------------+------------------------------------------------+\n| <v0.4             | v1.6             | v1.4               | v2.7                                           |\n+-------------------+------------------+--------------------+------------------------------------------------+\n\nThese are the unittests variants. See also: `/.travis.yml <https://github.com/jedie/django-reversion-compare/blob/master/.travis.yml>`_\nMaybe other versions are compatible, too.\n\n---------\nChangelog\n---------\n\n* *dev* `compare v0.10.0...master <https://github.com/jedie/django-reversion-compare/compare/v0.10.0...master>`_ \n\n    * TBC\n\n* v0.10.0 - 19.02.2020 `compare v0.9.1...v0.10.0 <https://github.com/jedie/django-reversion-compare/compare/v0.9.1...v0.10.0>`_ \n\n    * less restricted dependency specification see: `issues #120 <https://github.com/jedie/django-reversion-compare/issues/120>`_\n\n    * run tests with latest django-reversion version (currently v3.x)\n\n* v0.9.1 - 16.02.2020 `compare v0.9.0...v0.9.1 <https://github.com/jedie/django-reversion-compare/compare/v0.9.0...v0.9.1>`_ \n\n    * Modernize project setup and use poetry\n\n    * Apply pyupgrade and fix/update some f-strings\n\n    * Update test project\n\n* v0.9.0 - 19.01.2020 `compare v0.8.7...v0.9.0 <https://github.com/jedie/django-reversion-compare/compare/v0.8.7...v0.9.0>`_ \n\n    * Test with Python 3.8 and Django 3.0, too.\n\n    * Run tests via github actions, too.\n\n    * Remove support for Python 3.5 and Django v1.11\n\n    * `actually check if model is registered #115 <https://github.com/jedie/django-reversion-compare/pull/115>`_ contributed by willtho89\n\n    * `Remove python2 compatibility decorators #113 <https://github.com/jedie/django-reversion-compare/pull/113>`_ contributed by jeremy-engel\n\n    * `Show username and full name from custom user model #112 <https://github.com/jedie/django-reversion-compare/pull/112>`_ contributed by berekuk\n\n    * `Fix django-suit NoneType is not iterable #111 <https://github.com/jedie/django-reversion-compare/pull/111>`_ contributed by creativequality\n\n    * convert old format to f-strings via flynt\n\n    * Code style:\n\n        * sort imports with isort\n\n        * apply autopep8\n\n        * lint code in CI with flake8, isort and flynt\n\n* v0.8.7 - 06.01.2020 `compare v0.8.6...v0.8.7 <https://github.com/jedie/django-reversion-compare/compare/v0.8.6...v0.8.7>`_ \n\n    * Add new optional settings ``REVERSION_COMPARE_IGNORE_NOT_REGISTERED``, see: `issues #103 <https://github.com/jedie/django-reversion-compare/issues/103>`_\n\n    * reformat code with \'black\'\n\n    * some code cleanup\n\n* v0.8.6 - 04.01.2019 `compare v0.8.5...v0.8.6 <https://github.com/jedie/django-reversion-compare/compare/v0.8.5...v0.8.6>`_ \n\n    * Bugfix: `Use ".pk" instead of ".id" when referring to related object. <https://github.com/jedie/django-reversion-compare/pull/110>`_ contributed by `Peter Lisák <https://github.com/peterlisak>`_\n\n    * Run tests: Skip Django v1.8 and add Python v3.7\n\n* v0.8.5 - 13.09.2018 `compare v0.8.4...v0.8.5 <https://github.com/jedie/django-reversion-compare/compare/v0.8.4...v0.8.5>`_ \n\n    * `speed up delete checking <https://github.com/jedie/django-reversion-compare/pull/106>`_ contributed by `LegoStormtroopr <https://github.com/LegoStormtroopr>`_\n\n* v0.8.4 - 15.03.2018 `compare v0.8.3...v0.8.4 <https://github.com/jedie/django-reversion-compare/compare/v0.8.3...v0.8.4>`_ \n\n    * `Add Django 2.0 compatibility <https://github.com/jedie/django-reversion-compare/pull/102>`_ contributed by `samifahed <https://github.com/samifahed>`_\n\n* v0.8.3 - 21.12.2017 `compare v0.8.2...v0.8.3 <https://github.com/jedie/django-reversion-compare/compare/v0.8.2...v0.8.3>`_ \n\n    * refactor travis/tox/pytest/coverage stuff\n\n    * Tests can be run via ``python3 setup.py tox`` and/or ``python3 setup.py test``\n\n    * Test also with pypy3 on Travis CI.\n\n* `v0.8.2 - 06.12.2017 <https://github.com/jedie/django-reversion-compare/compare/v0.8.1...v0.8.2>`_:\n\n    * `Change ForeignKey relation compare <https://github.com/jedie/django-reversion-compare/pull/100>`_ contributed by `alaruss <https://github.com/alaruss>`_\n\n    * `Work around a type error triggered by taggit <https://github.com/jedie/django-reversion-compare/pull/86>`_ contributed by `Athemis <https://github.com/Athemis>`_\n\n    * minor code changes\n\n* `v0.8.1 - 02.10.2017 <https://github.com/jedie/django-reversion-compare/compare/v0.8.0...v0.8.1>`_:\n\n    * `Add added polish translation <https://github.com/jedie/django-reversion-compare/pull/99>`_ contributed by `w4rri0r3k <https://github.com/w4rri0r3k>`_\n\n    * Bugfix "Django>=1.11" in setup.py\n\n* `v0.8.0 - 17.08.2017 <https://github.com/jedie/django-reversion-compare/compare/v0.7.5...v0.8.0>`_:\n\n    * Run tests with Django v1.11 and drop tests with Django v1.9\n\n* `v0.7.5 - 24.04.2017 <https://github.com/jedie/django-reversion-compare/compare/v0.7.4...v0.7.5>`_:\n\n    * `Using the \'render\' function to ensure the execution of context processors properly <https://github.com/jedie/django-reversion-compare/pull/90>`_ contributed by `Rodrigo Pinheiro Marques de Araújo <https://github.com/fenrrir>`_\n\n* `v0.7.4 - 10.04.2017 <https://github.com/jedie/django-reversion-compare/compare/v0.7.3...v0.7.4>`_:\n\n    * Bugfix for Python 2: `compare unicode instead of bytes <https://github.com/jedie/django-reversion-compare/issues/89>`_ contributed by `Maksim Iakovlev <https://github.com/lampslave>`_\n\n    * `remove \'Django20Warning\' <https://github.com/jedie/django-reversion-compare/pull/88>`_ contributed by `Hugo Tácito <https://github.com/hugotacito>`_\n\n    * `Add \'Finnish\' localisations <https://github.com/jedie/django-reversion-compare/pull/87>`_ contributed by `Olli-Pekka Puolitaival <https://github.com/OPpuolitaival>`_\n\n* `v0.7.3 - 08.02.2017 <https://github.com/jedie/django-reversion-compare/compare/v0.7.2...v0.7.3>`_:\n\n    * `Fix case when model has template field which is ForeignKey <https://github.com/jedie/django-reversion-compare/pull/85>`_ contributed by `Lagovas <https://github.com/Lagovas>`_\n\n* `v0.7.2 - 20.10.2016 <https://github.com/jedie/django-reversion-compare/compare/v0.7.1...v0.7.2>`_:\n\n    * Add Django v1.10 support\n\n* `v0.7.1 - 29.08.2016 <https://github.com/jedie/django-reversion-compare/compare/v0.7.0...v0.7.1>`_:\n\n    * `Fix #79: missing import if **ADD_REVERSION_ADMIN != True** <https://github.com/jedie/django-reversion-compare/issues/79>`_\n\n* `v0.7.0 - 25.08.2016 <https://github.com/jedie/django-reversion-compare/compare/v0.6.3...v0.7.0>`_:\n\n    * `support only django-reversion >= 2.0 <https://github.com/jedie/django-reversion-compare/pull/76>`_ based on a contribution by `mshannon1123 <https://github.com/jedie/django-reversion-compare/pull/73>`_\n\n    * remove internal **reversion_api**\n\n    * Use tox\n\n* `v0.6.3 - 14.06.2016 <https://github.com/jedie/django-reversion-compare/compare/v0.6.2...v0.6.3>`_:\n\n    * `Remove unused and deprecated patters <https://github.com/jedie/django-reversion-compare/pull/69>`_ contributed by `codingjoe <https://github.com/codingjoe>`_\n\n    * `Fix django 1.10 warning #66 <https://github.com/jedie/django-reversion-compare/pull/66>`_ contributed by `pypetey <https://github.com/pypetey>`_\n\n* `v0.6.2 - 27.04.2016 <https://github.com/jedie/django-reversion-compare/compare/v0.6.1...v0.6.2>`_:\n\n    * `Added choices field representation #63 <https://github.com/jedie/django-reversion-compare/pull/63>`_ contributed by `amureki <https://github.com/amureki>`_\n\n    * `Check if related model has an integer as pk for ManyToMany fields. #64 <https://github.com/jedie/django-reversion-compare/pull/64>`_ contributed by `logaritmisk <https://github.com/logaritmisk>`_\n\n* `v0.6.1 - 16.02.2016 <https://github.com/jedie/django-reversion-compare/compare/v0.6.0...v0.6.1>`_:\n\n    * `pull #61 <https://github.com/jedie/django-reversion-compare/pull/61>`_: Fix error when ManyToMany relations didn\'t exist contributed by `Diederik van der Boor <https://github.com/vdboor>`_\n\n* `v0.6.0 - 03.02.2016 <https://github.com/jedie/django-reversion-compare/compare/v0.5.6...v0.6.0>`_:\n\n    * Added Dutch translation contributed by `Sae X <https://github.com/SaeX>`_\n\n    * Add support for Django 1.9\n\n    * Nicer boolean compare: `#57 <https://github.com/jedie/django-reversion-compare/issues/57>`_\n\n    * Fix `#58 compare followed reverse foreign relation fields that are on a non-abstract parent class <https://github.com/jedie/django-reversion-compare/issues/58>`_ contributed by LegoStormtroopr\n\n* `v0.5.6 - 23.09.2015 <https://github.com/jedie/django-reversion-compare/compare/v0.5.5...v0.5.6>`_:\n\n    * NEW: Class-Based-View to create non-admin views and greek translation contributed by `Serafeim Papastefanos <https://github.com/spapas>`_.\n\n* `v0.5.5 - 24.07.2015 <https://github.com/jedie/django-reversion-compare/compare/v0.5.4...v0.5.5>`_:\n\n    * UnboundLocalError (\'version\') when creating deleted list in get_many_to_something() `#41 <https://github.com/jedie/django-reversion-compare/pull/41>`_\n\n* `v0.5.4 - 22.07.2015 <https://github.com/jedie/django-reversion-compare/compare/v0.5.3...v0.5.4>`_:\n\n    * One to one field custom related name fix `#42 <https://github.com/jedie/django-reversion-compare/pull/42>`_ (contributed by frwickst and aemdy)\n\n* `v0.5.3 - 13.07.2015 <https://github.com/jedie/django-reversion-compare/compare/v0.5.2...v0.5.3>`_:\n\n    * Update admin.py to avoid RemovedInDjango19Warning (contributed by luzfcb)\n\n* `v0.5.2 - 14.04.2015 <https://github.com/jedie/django-reversion-compare/compare/v0.5.1...v0.5.2>`_:\n\n    * contributed by Samuel Spencer:\n\n        * Added Django 1.8 support: `pull #35 <https://github.com/jedie/django-reversion-compare/pull/35>`_\n\n        * list of changes for reverse fields incorrectly includes a "deletion" for the item that was added in: `issues #34 <https://github.com/jedie/django-reversion-compare/issues/34>`_\n\n* `v0.5.1 - 28.02.2015 <https://github.com/jedie/django-reversion-compare/compare/v0.5.0...v0.5.1>`_:\n\n    * activate previous/next links and add unitests for them\n\n* `v0.5.0 - 27.02.2015 <https://github.com/jedie/django-reversion-compare/compare/v0.4.0...v0.5.0>`_:\n\n    * refactory unittests, test with Django v1.7 and Python 2.7 & 3.4\n\n* `v0.4.0 - 02.02.2015 <https://github.com/jedie/django-reversion-compare/compare/v0.3.5...v0.4.0>`_:\n\n    * Updates for django 1.7 support\n\n    * Add ``settings.ADD_REVERSION_ADMIN``\n\n* v0.3.5 - 03.01.2013:\n\n    * Remove date from version string. `issues 9 <https://github.com/jedie/django-reversion-compare/issues/9>`_\n\n* v0.3.4 - 20.06.2012:\n\n    * Use VersionAdmin.revision_manager rather than default_revision_manager, contributed by Mark Lavin - see: `pull request 7 <https://github.com/jedie/django-reversion-compare/pull/7>`_\n\n    * Use logging for all debug prints, contributed by Bojan Mihelac - see: `pull request 8 <https://github.com/jedie/django-reversion-compare/pull/8>`_\n\n* v0.3.3 - 11.06.2012:\n\n    * Bugfix "ValueError: zero length field name in format" with Python 2.6 `issues 5 <https://github.com/jedie/django-reversion-compare/issues/5>`_\n\n* v0.3.2 - 04.06.2012:\n\n    * Bugfix for Python 2.6 in unified_diff(), see: `AttributeError: \'module\' object has no attribute \'_format_range_unified\' <https://github.com/jedie/django-reversion-compare/issues/5>`_\n\n* v0.3.1 - 01.06.2012:\n\n    * Bugfix: force unicode in html diff\n\n    * Bugfix in unittests\n\n* v0.3.0 - 16.05.2012:\n\n    * Enhanced handling of m2m changes with follow and non-follow relations.\n\n* v0.2.2 - 15.05.2012:\n\n    * Compare many-to-many in the right way.\n\n* v0.2.1 - 10.05.2012:\n\n    * Bugfix for models which has no m2m field: `https://github.com/jedie/django-reversion-compare/commit/c8e042945a6e78e5540b6ae27666f9b0cfc94880 <https://github.com/jedie/django-reversion-compare/commit/c8e042945a6e78e5540b6ae27666f9b0cfc94880>`_\n\n* v0.2.0 - 09.05.2012:\n\n    * many-to-many compare works, too.\n\n* v0.1.0 - 08.05.2012:\n\n    * First release\n\n* v0.0.1 - 08.05.2012:\n\n    * collect all compare stuff from old "diff" branch\n\n    * see also: `https://github.com/etianen/django-reversion/issues/147 <https://github.com/etianen/django-reversion/issues/147>`_\n\n-----\nLinks\n-----\n\n+-----------------+-------------------------------------------------------+\n| Github          | `https://github.com/jedie/django-reversion-compare`_  |\n+-----------------+-------------------------------------------------------+\n| Python Packages | `https://pypi.org/project/django-reversion-compare/`_ |\n+-----------------+-------------------------------------------------------+\n\n.. _https://github.com/jedie/django-reversion-compare: https://github.com/jedie/django-reversion-compare\n.. _https://pypi.org/project/django-reversion-compare/: https://pypi.org/project/django-reversion-compare/\n\n--------\nDonation\n--------\n\n* `paypal.me/JensDiemer <https://www.paypal.me/JensDiemer>`_\n\n* `Flattr This! <https://flattr.com/submit/auto?uid=jedie&url=https%3A%2F%2Fgithub.com%2Fjedie%2Fdjango-reversion-compare%2F>`_\n\n* Send `Bitcoins <https://www.bitcoin.org/>`_ to `1823RZ5Md1Q2X5aSXRC5LRPcYdveCiVX6F <https://blockexplorer.com/address/1823RZ5Md1Q2X5aSXRC5LRPcYdveCiVX6F>`_\n\n------------\n\n``Note: this file is generated from README.creole 2020-02-19 09:39:16 with "python-creole"``',
    'author': 'Jens Diemer',
    'author_email': 'django-reversion-compare@jensdiemer.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jedie/django-reversion-compare/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
    admin
    ~~~~~

    Admin extensions for django-reversion-compare

    :copyleft: 2012-2019 by the django-reversion-compare team, see AUTHORS for more details.
    :license: GNU GPL v3 or above, see LICENSE for more details.
"""


import logging

from django.conf import settings
from django.conf.urls import url
from django.contrib import admin
from django.contrib.admin.utils import quote, unquote
from django.http import Http404
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.text import capfirst
from django.utils.translation import ugettext as _

from reversion.admin import VersionAdmin
from reversion.models import Revision, Version
from reversion_compare.forms import SelectDiffForm
from reversion_compare.mixins import CompareMethodsMixin, CompareMixin

logger = logging.getLogger(__name__)


class BaseCompareVersionAdmin(CompareMixin, VersionAdmin):
    """
    Enhanced version of VersionAdmin with a flexible compare version API.

    You can define own method to compare fields in two ways (in this order):

        Create a method for a field via the field name, e.g.:
            "compare_%s" % field_name

        Create a method for every field by his internal type
            "compare_%s" % field.get_internal_type()

        see: https://docs.djangoproject.com/en/1.4/howto/custom-model-fields/#django.db.models.Field.get_internal_type

    If no method defined it would build a simple ndiff from repr().

    example:

    ----------------------------------------------------------------------------
    class MyModel(models.Model):
        date_created = models.DateTimeField(auto_now_add=True)
        last_update = models.DateTimeField(auto_now=True)
        user = models.ForeignKey(User)
        content = models.TextField()
        sub_text = models.ForeignKey(FooBar)

    class MyModelAdmin(CompareVersionAdmin):
        def compare_DateTimeField(self, obj, version1, version2, value1, value2):
            ''' compare all model datetime model field in ISO format '''
            date1 = value1.isoformat(" ")
            date2 = value2.isoformat(" ")
            html = html_diff(date1, date2)
            return html

        def compare_sub_text(self, obj, version1, version2, value1, value2):
            ''' field_name example '''
            return "%s -> %s" % (value1, value2)

    ----------------------------------------------------------------------------
    """

    # Template file used for the compare view:
    compare_template = "reversion-compare/compare.html"

    # change template from django-reversion to add compare selection form:
    object_history_template = "reversion-compare/object_history.html"

    def get_urls(self):
        """Returns the additional urls used by the Reversion admin."""
        urls = super().get_urls()
        admin_site = self.admin_site
        opts = self.model._meta
        info = opts.app_label, opts.model_name
        reversion_urls = [
            url("^([^/]+)/history/compare/$", admin_site.admin_view(self.compare_view), name="%s_%s_compare" % info)
        ]
        return reversion_urls + urls

    def _get_action_list(self, request, object_id, extra_context=None):
        """Renders the history view."""
        object_id = unquote(object_id)  # Underscores in primary key get quoted to "_5F"
        opts = self.model._meta
        action_list = [
            {
                "version": version,
                "revision": version.revision,
                "url": reverse(
                    f"{self.admin_site.name}:{opts.app_label}_{opts.model_name}_revision",
                    args=(quote(version.object_id), version.id),
                ),
            }
            for version in self._order_version_queryset(
                Version.objects.get_for_object_reference(self.model, object_id).select_related("revision__user")
            )
        ]
        return action_list

    def history_view(self, request, object_id, extra_context=None):
        """Renders the history view."""
        action_list = self._get_action_list(request, object_id, extra_context=extra_context)

        if len(action_list) < 2:
            # Less than two history items aren't enough to compare ;)
            comparable = False
        else:
            comparable = True
            # for pre selecting the compare radio buttons depend on the ordering:
            if self.history_latest_first:
                action_list[0]["first"] = True
                action_list[1]["second"] = True
            else:
                action_list[-1]["first"] = True
                action_list[-2]["second"] = True

        # Compile the context.
        context = {"action_list": action_list, "comparable": comparable, "compare_view": True}
        context.update(extra_context or {})
        return super().history_view(request, object_id, context)

    def compare_view(self, request, object_id, extra_context=None):
        """
        compare two versions.
        Used self.make_compare() to create the html diff.
        """
        if self.compare is None:
            raise Http404("Compare view not enabled.")

        form = SelectDiffForm(request.GET)
        if not form.is_valid():
            msg = "Wrong version IDs."
            if settings.DEBUG:
                msg += " (form errors: %s)" % ", ".join(form.errors)
            raise Http404(msg)

        version_id1 = form.cleaned_data["version_id1"]
        version_id2 = form.cleaned_data["version_id2"]

        if version_id1 > version_id2:
            # Compare always the newest one (#2) with the older one (#1)
            version_id1, version_id2 = version_id2, version_id1

        object_id = unquote(object_id)  # Underscores in primary key get quoted to "_5F"
        obj = get_object_or_404(self.model, pk=object_id)
        queryset = Version.objects.get_for_object(obj)
        version1 = get_object_or_404(queryset, pk=version_id1)
        version2 = get_object_or_404(queryset, pk=version_id2)

        next_version = queryset.filter(pk__gt=version_id2).last()
        prev_version = queryset.filter(pk__lt=version_id1).first()

        compare_data, has_unfollowed_fields = self.compare(obj, version1, version2)

        opts = self.model._meta

        context = {
            **self.admin_site.each_context(request),
            "opts": opts,
            "app_label": opts.app_label,
            "model_name": capfirst(opts.verbose_name),
            "title": _("Compare %(name)s") % {"name": version1.object_repr},
            "obj": obj,
            "compare_data": compare_data,
            "has_unfollowed_fields": has_unfollowed_fields,
            "version1": version1,
            "version2": version2,
            "changelist_url": reverse(f"{self.admin_site.name}:{opts.app_label}_{opts.model_name}_changelist"),
            "change_url": reverse(
                f"{self.admin_site.name}:{opts.app_label}_{opts.model_name}_change", args=(quote(obj.pk),)
            ),
            "original": obj,
            "history_url": reverse(
                f"{self.admin_site.name}:{opts.app_label}_{opts.model_name}_history", args=(quote(obj.pk),)
            ),
        }

        # don't use urlencode with dict for generate prev/next-urls
        # Otherwise we can't unitests it!
        if next_version:
            next_url = f"?version_id1={version2.id:d}&version_id2={next_version.id:d}"
            context.update({"next_url": next_url})
        if prev_version:
            prev_url = f"?version_id1={prev_version.id:d}&version_id2={version1.id:d}"
            context.update({"prev_url": prev_url})

        context.update(extra_context or {})
        return render(request, self.compare_template or self._get_template_list("compare.html"), context)


class CompareVersionAdmin(CompareMethodsMixin, BaseCompareVersionAdmin):
    """
    expand the base class with prepared compare methods. This is the
    class to inherit
    """

    pass


if hasattr(settings, "ADD_REVERSION_ADMIN") and settings.ADD_REVERSION_ADMIN:

    class RevisionAdmin(admin.ModelAdmin):
        list_display = ("id", "date_created", "user", "comment")
        list_display_links = ("date_created",)
        date_hierarchy = "date_created"
        ordering = ("-date_created",)
        list_filter = ("user", "comment")
        search_fields = ("user", "comment")

    admin.site.register(Revision, RevisionAdmin)

    class VersionAdmin(admin.ModelAdmin):
        list_display = ("object_repr", "revision", "object_id", "content_type", "format")
        list_display_links = ("object_repr", "object_id")
        list_filter = ("content_type", "format")
        search_fields = ("object_repr", "serialized_data")

    admin.site.register(Version, VersionAdmin)
